@@ -1,22 +1,25 @@
 use std::mem::size_of;
-use std::os::raw::{
-    c_char, c_int, c_long, c_longlong, c_short, c_uint, c_ulong, c_ulonglong, c_ushort,
-};
+use std::os::raw::{c_int, c_long, c_longlong, c_short, c_uint, c_ulong, c_ulonglong, c_ushort};
 use std::ptr;
 
+#[cfg(feature = "half")]
+use half::{bf16, f16};
 use num_traits::{Bounded, Zero};
+#[cfg(feature = "half")]
+use pyo3::sync::GILOnceCell;
 use pyo3::{
+    conversion::IntoPyObject,
     exceptions::{PyIndexError, PyValueError},
     ffi::{self, PyTuple_Size},
-    pyobject_native_type_extract, pyobject_native_type_named,
-    types::{PyDict, PyTuple, PyType},
-    AsPyPointer, FromPyObject, FromPyPointer, IntoPyPointer, PyAny, PyNativeType, PyObject,
-    PyResult, PyTypeInfo, Python, ToPyObject,
+    pyobject_native_type_named,
+    types::{PyAnyMethods, PyDict, PyDictMethods, PyTuple, PyType},
+    Borrowed, Bound, Py, PyAny, PyObject, PyResult, PyTypeInfo, Python,
 };
 
 use crate::npyffi::{
-    NpyTypes, PyArray_Descr, NPY_ALIGNED_STRUCT, NPY_BYTEORDER_CHAR, NPY_ITEM_HASOBJECT, NPY_TYPES,
-    PY_ARRAY_API,
+    NpyTypes, PyArray_Descr, PyDataType_ALIGNMENT, PyDataType_ELSIZE, PyDataType_FIELDS,
+    PyDataType_FLAGS, PyDataType_NAMES, PyDataType_SUBARRAY, NPY_ALIGNED_STRUCT,
+    NPY_BYTEORDER_CHAR, NPY_ITEM_HASOBJECT, NPY_TYPES, PY_ARRAY_API,
 };
 
 pub use num_complex::{Complex32, Complex64};
@@ -26,55 +29,52 @@ pub use num_complex::{Complex32, Complex64};
 /// # Example
 ///
 /// ```
-/// use numpy::{dtype, get_array_module, PyArrayDescr};
-/// use numpy::pyo3::{types::IntoPyDict, Python};
+/// use numpy::{dtype, get_array_module, PyArrayDescr, PyArrayDescrMethods};
+/// use numpy::pyo3::{types::{IntoPyDict, PyAnyMethods}, Python, ffi::c_str};
 ///
+/// # fn main() -> pyo3::PyResult<()> {
 /// Python::with_gil(|py| {
-///     let locals = [("np", get_array_module(py).unwrap())].into_py_dict(py);
+///     let locals = [("np", get_array_module(py)?)].into_py_dict(py)?;
 ///
-///     let dt: &PyArrayDescr = py
-///         .eval("np.array([1, 2, 3.0]).dtype", Some(locals), None)
-///         .unwrap()
-///         .downcast()
-///         .unwrap();
+///     let dt = py
+///         .eval(c_str!("np.array([1, 2, 3.0]).dtype"), Some(&locals), None)?
+///         .downcast_into::<PyArrayDescr>()?;
 ///
-///     assert!(dt.is_equiv_to(dtype::<f64>(py)));
-/// });
+///     assert!(dt.is_equiv_to(&dtype::<f64>(py)));
+/// #   Ok(())
+/// })
+/// # }
 /// ```
 ///
 /// [dtype]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.html
+#[repr(transparent)]
 pub struct PyArrayDescr(PyAny);
 
 pyobject_native_type_named!(PyArrayDescr);
 
 unsafe impl PyTypeInfo for PyArrayDescr {
-    type AsRefTarget = Self;
-
     const NAME: &'static str = "PyArrayDescr";
     const MODULE: Option<&'static str> = Some("numpy");
 
     #[inline]
-    fn type_object_raw(py: Python) -> *mut ffi::PyTypeObject {
+    fn type_object_raw<'py>(py: Python<'py>) -> *mut ffi::PyTypeObject {
         unsafe { PY_ARRAY_API.get_type_object(py, NpyTypes::PyArrayDescr_Type) }
     }
-
-    fn is_type_of(ob: &PyAny) -> bool {
-        unsafe {
-            ffi::PyObject_TypeCheck(
-                ob.as_ptr(),
-                PY_ARRAY_API.get_type_object(ob.py(), NpyTypes::PyArrayDescr_Type),
-            ) > 0
-        }
-    }
 }
 
-pyobject_native_type_extract!(PyArrayDescr);
-
 /// Returns the type descriptor ("dtype") for a registered type.
-pub fn dtype<T: Element>(py: Python) -> &PyArrayDescr {
+#[inline]
+pub fn dtype<'py, T: Element>(py: Python<'py>) -> Bound<'py, PyArrayDescr> {
     T::get_dtype(py)
 }
 
+/// Deprecated name for [`dtype`].
+#[deprecated(since = "0.23.0", note = "renamed to `dtype`")]
+#[inline]
+pub fn dtype_bound<'py, T: Element>(py: Python<'py>) -> Bound<'py, PyArrayDescr> {
+    dtype::<T>(py)
+}
+
 impl PyArrayDescr {
     /// Creates a new type descriptor ("dtype") object from an arbitrary object.
     ///
@@ -82,55 +82,97 @@ impl PyArrayDescr {
     ///
     /// [dtype]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.html
     #[inline]
-    pub fn new<'py, T: ToPyObject + ?Sized>(py: Python<'py>, obj: &T) -> PyResult<&'py Self> {
-        Self::new_impl(py, obj.to_object(py))
+    pub fn new<'a, 'py, T>(py: Python<'py>, ob: T) -> PyResult<Bound<'py, Self>>
+    where
+        T: IntoPyObject<'py>,
+    {
+        fn inner<'py>(
+            py: Python<'py>,
+            obj: Borrowed<'_, 'py, PyAny>,
+        ) -> PyResult<Bound<'py, PyArrayDescr>> {
+            let mut descr: *mut PyArray_Descr = ptr::null_mut();
+            unsafe {
+                // None is an invalid input here and is not converted to NPY_DEFAULT_TYPE
+                PY_ARRAY_API.PyArray_DescrConverter2(py, obj.as_ptr(), &mut descr);
+                Bound::from_owned_ptr_or_err(py, descr.cast())
+                    .map(|any| any.downcast_into_unchecked())
+            }
+        }
+
+        inner(
+            py,
+            ob.into_pyobject(py)
+                .map_err(Into::into)?
+                .into_any()
+                .as_borrowed(),
+        )
     }
 
-    fn new_impl<'py>(py: Python<'py>, obj: PyObject) -> PyResult<&'py Self> {
-        let mut descr: *mut PyArray_Descr = ptr::null_mut();
+    /// Deprecated name for [`PyArrayDescr::new`].
+    #[deprecated(since = "0.23.0", note = "renamed to `PyArrayDescr::new`")]
+    #[allow(deprecated)]
+    #[inline]
+    pub fn new_bound<'py, T: pyo3::ToPyObject + ?Sized>(
+        py: Python<'py>,
+        ob: &T,
+    ) -> PyResult<Bound<'py, Self>> {
+        Self::new(py, ob.to_object(py))
+    }
+
+    /// Shortcut for creating a type descriptor of `object` type.
+    #[inline]
+    pub fn object(py: Python<'_>) -> Bound<'_, Self> {
+        Self::from_npy_type(py, NPY_TYPES::NPY_OBJECT)
+    }
+
+    /// Deprecated name for [`PyArrayDescr::object`].
+    #[deprecated(since = "0.23.0", note = "renamed to `PyArrayDescr::object`")]
+    #[inline]
+    pub fn object_bound(py: Python<'_>) -> Bound<'_, Self> {
+        Self::object(py)
+    }
+
+    /// Returns the type descriptor for a registered type.
+    #[inline]
+    pub fn of<'py, T: Element>(py: Python<'py>) -> Bound<'py, Self> {
+        T::get_dtype(py)
+    }
+
+    /// Deprecated name for [`PyArrayDescr::of`].
+    #[deprecated(since = "0.23.0", note = "renamed to `PyArrayDescr::of`")]
+    #[inline]
+    pub fn of_bound<'py, T: Element>(py: Python<'py>) -> Bound<'py, Self> {
+        Self::of::<T>(py)
+    }
+
+    fn from_npy_type<'py>(py: Python<'py>, npy_type: NPY_TYPES) -> Bound<'py, Self> {
         unsafe {
-            // None is an invalid input here and is not converted to NPY_DEFAULT_TYPE
-            PY_ARRAY_API.PyArray_DescrConverter2(py, obj.as_ptr(), &mut descr as *mut _);
-            py.from_owned_ptr_or_err(descr as _)
+            let descr = PY_ARRAY_API.PyArray_DescrFromType(py, npy_type as _);
+            Bound::from_owned_ptr(py, descr.cast()).downcast_into_unchecked()
         }
     }
 
-    /// Returns `self` as `*mut PyArray_Descr`.
-    pub fn as_dtype_ptr(&self) -> *mut PyArray_Descr {
-        self.as_ptr() as _
+    pub(crate) fn new_from_npy_type<'py>(py: Python<'py>, npy_type: NPY_TYPES) -> Bound<'py, Self> {
+        unsafe {
+            let descr = PY_ARRAY_API.PyArray_DescrNewFromType(py, npy_type as _);
+            Bound::from_owned_ptr(py, descr.cast()).downcast_into_unchecked()
+        }
     }
+}
+
+/// Implementation of functionality for [`PyArrayDescr`].
+#[doc(alias = "PyArrayDescr")]
+pub trait PyArrayDescrMethods<'py>: Sealed {
+    /// Returns `self` as `*mut PyArray_Descr`.
+    fn as_dtype_ptr(&self) -> *mut PyArray_Descr;
 
     /// Returns `self` as `*mut PyArray_Descr` while increasing the reference count.
     ///
     /// Useful in cases where the descriptor is stolen by the API.
-    pub fn into_dtype_ptr(&self) -> *mut PyArray_Descr {
-        self.into_ptr() as _
-    }
-
-    /// Shortcut for creating a type descriptor of `object` type.
-    pub fn object(py: Python) -> &Self {
-        Self::from_npy_type(py, NPY_TYPES::NPY_OBJECT)
-    }
-
-    /// Returns the type descriptor for a registered type.
-    pub fn of<T: Element>(py: Python) -> &Self {
-        T::get_dtype(py)
-    }
+    fn into_dtype_ptr(self) -> *mut PyArray_Descr;
 
     /// Returns true if two type descriptors are equivalent.
-    pub fn is_equiv_to(&self, other: &Self) -> bool {
-        unsafe {
-            PY_ARRAY_API.PyArray_EquivTypes(self.py(), self.as_dtype_ptr(), other.as_dtype_ptr())
-                != 0
-        }
-    }
-
-    fn from_npy_type(py: Python, npy_type: NPY_TYPES) -> &Self {
-        unsafe {
-            let descr = PY_ARRAY_API.PyArray_DescrFromType(py, npy_type as _);
-            py.from_owned_ptr(descr as _)
-        }
-    }
+    fn is_equiv_to(&self, other: &Self) -> bool;
 
     /// Returns the [array scalar][arrays-scalars] corresponding to this type descriptor.
     ///
@@ -138,16 +180,7 @@ impl PyArrayDescr {
     ///
     /// [arrays-scalars]: https://numpy.org/doc/stable/reference/arrays.scalars.html
     /// [dtype-type]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.type.html
-    pub fn typeobj(&self) -> &PyType {
-        let dtype_type_ptr = unsafe { *self.as_dtype_ptr() }.typeobj;
-        unsafe { PyType::from_type_ptr(self.py(), dtype_type_ptr) }
-    }
-
-    #[doc(hidden)]
-    #[deprecated(note = "`get_type()` is deprecated, please use `typeobj()` instead")]
-    pub fn get_type(&self) -> &PyType {
-        self.typeobj()
-    }
+    fn typeobj(&self) -> Bound<'py, PyType>;
 
     /// Returns a unique number for each of the 21 different built-in
     /// [enumerated types][enumerated-types].
@@ -158,8 +191,8 @@ impl PyArrayDescr {
     ///
     /// [enumerated-types]: https://numpy.org/doc/stable/reference/c-api/dtype.html#enumerated-types
     /// [dtype-num]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.num.html
-    pub fn num(&self) -> c_int {
-        unsafe { *self.as_dtype_ptr() }.type_num
+    fn num(&self) -> c_int {
+        unsafe { &*self.as_dtype_ptr() }.type_num
     }
 
     /// Returns the element size of this type descriptor.
@@ -167,18 +200,14 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.itemsize`][dtype-itemsize].
     ///
     /// [dtype-itemsiize]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.itemsize.html
-    pub fn itemsize(&self) -> usize {
-        unsafe { *self.as_dtype_ptr() }.elsize.max(0) as _
-    }
+    fn itemsize(&self) -> usize;
 
     /// Returns the required alignment (bytes) of this type descriptor according to the compiler.
     ///
     /// Equivalent to [`numpy.dtype.alignment`][dtype-alignment].
     ///
     /// [dtype-alignment]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.alignment.html
-    pub fn alignment(&self) -> usize {
-        unsafe { *self.as_dtype_ptr() }.alignment.max(0) as _
-    }
+    fn alignment(&self) -> usize;
 
     /// Returns an ASCII character indicating the byte-order of this type descriptor object.
     ///
@@ -187,8 +216,8 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.byteorder`][dtype-byteorder].
     ///
     /// [dtype-byteorder]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.byteorder.html
-    pub fn byteorder(&self) -> u8 {
-        unsafe { *self.as_dtype_ptr() }.byteorder.max(0) as _
+    fn byteorder(&self) -> u8 {
+        unsafe { &*self.as_dtype_ptr() }.byteorder.max(0) as _
     }
 
     /// Returns a unique ASCII character for each of the 21 different built-in types.
@@ -198,8 +227,8 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.char`][dtype-char].
     ///
     /// [dtype-char]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.char.html
-    pub fn char(&self) -> u8 {
-        unsafe { *self.as_dtype_ptr() }.type_.max(0) as _
+    fn char(&self) -> u8 {
+        unsafe { &*self.as_dtype_ptr() }.type_.max(0) as _
     }
 
     /// Returns an ASCII character (one of `biufcmMOSUV`) identifying the general kind of data.
@@ -209,8 +238,8 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.kind`][dtype-kind].
     ///
     /// [dtype-kind]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.kind.html
-    pub fn kind(&self) -> u8 {
-        unsafe { *self.as_dtype_ptr() }.kind.max(0) as _
+    fn kind(&self) -> u8 {
+        unsafe { &*self.as_dtype_ptr() }.kind.max(0) as _
     }
 
     /// Returns bit-flags describing how this type descriptor is to be interpreted.
@@ -218,21 +247,14 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.flags`][dtype-flags].
     ///
     /// [dtype-flags]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.flags.html
-    pub fn flags(&self) -> c_char {
-        unsafe { *self.as_dtype_ptr() }.flags
-    }
+    fn flags(&self) -> u64;
 
     /// Returns the number of dimensions if this type descriptor represents a sub-array, and zero otherwise.
     ///
     /// Equivalent to [`numpy.dtype.ndim`][dtype-ndim].
     ///
     /// [dtype-ndim]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.ndim.html
-    pub fn ndim(&self) -> usize {
-        if !self.has_subarray() {
-            return 0;
-        }
-        unsafe { PyTuple_Size((*((*self.as_dtype_ptr()).subarray)).shape).max(0) as _ }
-    }
+    fn ndim(&self) -> usize;
 
     /// Returns the type descriptor for the base element of subarrays, regardless of their dimension or shape.
     ///
@@ -241,15 +263,7 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.base`][dtype-base].
     ///
     /// [dtype-base]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.base.html
-    pub fn base(&self) -> &PyArrayDescr {
-        if !self.has_subarray() {
-            self
-        } else {
-            unsafe {
-                Self::from_borrowed_ptr(self.py(), (*(*self.as_dtype_ptr()).subarray).base as _)
-            }
-        }
-    }
+    fn base(&self) -> Bound<'py, PyArrayDescr>;
 
     /// Returns the shape of the sub-array.
     ///
@@ -258,25 +272,14 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.shape`][dtype-shape].
     ///
     /// [dtype-shape]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.shape.html
-    pub fn shape(&self) -> Vec<usize> {
-        if !self.has_subarray() {
-            Vec::new()
-        } else {
-            // NumPy guarantees that shape is a tuple of non-negative integers so this should never panic.
-            unsafe {
-                PyTuple::from_borrowed_ptr(self.py(), (*(*self.as_dtype_ptr()).subarray).shape)
-            }
-            .extract()
-            .unwrap()
-        }
-    }
+    fn shape(&self) -> Vec<usize>;
 
     /// Returns true if the type descriptor contains any reference-counted objects in any fields or sub-dtypes.
     ///
     /// Equivalent to [`numpy.dtype.hasobject`][dtype-hasobject].
     ///
     /// [dtype-hasobject]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.hasobject.html
-    pub fn has_object(&self) -> bool {
+    fn has_object(&self) -> bool {
         self.flags() & NPY_ITEM_HASOBJECT != 0
     }
 
@@ -288,24 +291,22 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.isalignedstruct`][dtype-isalignedstruct].
     ///
     /// [dtype-isalignedstruct]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.isalignedstruct.html
-    pub fn is_aligned_struct(&self) -> bool {
+    fn is_aligned_struct(&self) -> bool {
         self.flags() & NPY_ALIGNED_STRUCT != 0
     }
 
     /// Returns true if the type descriptor is a sub-array.
-    pub fn has_subarray(&self) -> bool {
-        // equivalent to PyDataType_HASSUBARRAY(self)
-        unsafe { !(*self.as_dtype_ptr()).subarray.is_null() }
-    }
+    ///
+    /// Equivalent to PyDataType_HASSUBARRAY(self).
+    fn has_subarray(&self) -> bool;
 
     /// Returns true if the type descriptor is a structured type.
-    pub fn has_fields(&self) -> bool {
-        // equivalent to PyDataType_HASFIELDS(self)
-        unsafe { !(*self.as_dtype_ptr()).names.is_null() }
-    }
+    ///
+    /// Equivalent to PyDataType_HASFIELDS(self).
+    fn has_fields(&self) -> bool;
 
     /// Returns true if type descriptor byteorder is native, or `None` if not applicable.
-    pub fn is_native_byteorder(&self) -> Option<bool> {
+    fn is_native_byteorder(&self) -> Option<bool> {
         // based on PyArray_ISNBO(self->byteorder)
         match self.byteorder() {
             b'=' => Some(true),
@@ -321,13 +322,7 @@ impl PyArrayDescr {
     /// Equivalent to [`numpy.dtype.names`][dtype-names].
     ///
     /// [dtype-names]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.names.html
-    pub fn names(&self) -> Option<Vec<&str>> {
-        if !self.has_fields() {
-            return None;
-        }
-        let names = unsafe { PyTuple::from_borrowed_ptr(self.py(), (*self.as_dtype_ptr()).names) };
-        FromPyObject::extract(names).ok()
-    }
+    fn names(&self) -> Option<Vec<String>>;
 
     /// Returns the type descriptor and offset of the field with the given name.
     ///
@@ -339,31 +334,142 @@ impl PyArrayDescr {
     /// Equivalent to retrieving a single item from [`numpy.dtype.fields`][dtype-fields].
     ///
     /// [dtype-fields]: https://numpy.org/doc/stable/reference/generated/numpy.dtype.fields.html
-    pub fn get_field(&self, name: &str) -> PyResult<(&PyArrayDescr, usize)> {
+    fn get_field(&self, name: &str) -> PyResult<(Bound<'py, PyArrayDescr>, usize)>;
+}
+
+mod sealed {
+    pub trait Sealed {}
+}
+
+use sealed::Sealed;
+
+impl<'py> PyArrayDescrMethods<'py> for Bound<'py, PyArrayDescr> {
+    fn as_dtype_ptr(&self) -> *mut PyArray_Descr {
+        self.as_ptr() as _
+    }
+
+    fn into_dtype_ptr(self) -> *mut PyArray_Descr {
+        self.into_ptr() as _
+    }
+
+    fn is_equiv_to(&self, other: &Self) -> bool {
+        let self_ptr = self.as_dtype_ptr();
+        let other_ptr = other.as_dtype_ptr();
+
+        unsafe {
+            self_ptr == other_ptr
+                || PY_ARRAY_API.PyArray_EquivTypes(self.py(), self_ptr, other_ptr) != 0
+        }
+    }
+
+    fn typeobj(&self) -> Bound<'py, PyType> {
+        let dtype_type_ptr = unsafe { &*self.as_dtype_ptr() }.typeobj;
+        unsafe { PyType::from_borrowed_type_ptr(self.py(), dtype_type_ptr) }
+    }
+
+    fn itemsize(&self) -> usize {
+        unsafe { PyDataType_ELSIZE(self.py(), self.as_dtype_ptr()).max(0) as _ }
+    }
+
+    fn alignment(&self) -> usize {
+        unsafe { PyDataType_ALIGNMENT(self.py(), self.as_dtype_ptr()).max(0) as _ }
+    }
+
+    fn flags(&self) -> u64 {
+        unsafe { PyDataType_FLAGS(self.py(), self.as_dtype_ptr()) as _ }
+    }
+
+    fn ndim(&self) -> usize {
+        let subarray = unsafe { PyDataType_SUBARRAY(self.py(), self.as_dtype_ptr()).as_ref() };
+        match subarray {
+            None => 0,
+            Some(subarray) => unsafe { PyTuple_Size(subarray.shape) }.max(0) as _,
+        }
+    }
+
+    fn base(&self) -> Bound<'py, PyArrayDescr> {
+        let subarray = unsafe { PyDataType_SUBARRAY(self.py(), self.as_dtype_ptr()).as_ref() };
+        match subarray {
+            None => self.clone(),
+            Some(subarray) => unsafe {
+                Bound::from_borrowed_ptr(self.py(), subarray.base.cast()).downcast_into_unchecked()
+            },
+        }
+    }
+
+    fn shape(&self) -> Vec<usize> {
+        let subarray = unsafe { PyDataType_SUBARRAY(self.py(), self.as_dtype_ptr()).as_ref() };
+        match subarray {
+            None => Vec::new(),
+            Some(subarray) => {
+                // NumPy guarantees that shape is a tuple of non-negative integers so this should never panic.
+                let shape = unsafe { Borrowed::from_ptr(self.py(), subarray.shape) };
+                shape.extract().unwrap()
+            }
+        }
+    }
+
+    fn has_subarray(&self) -> bool {
+        unsafe { !PyDataType_SUBARRAY(self.py(), self.as_dtype_ptr()).is_null() }
+    }
+
+    fn has_fields(&self) -> bool {
+        unsafe { !PyDataType_NAMES(self.py(), self.as_dtype_ptr()).is_null() }
+    }
+
+    fn names(&self) -> Option<Vec<String>> {
+        if !self.has_fields() {
+            return None;
+        }
+        let names = unsafe {
+            Borrowed::from_ptr(self.py(), PyDataType_NAMES(self.py(), self.as_dtype_ptr()))
+        };
+        names.extract().ok()
+    }
+
+    fn get_field(&self, name: &str) -> PyResult<(Bound<'py, PyArrayDescr>, usize)> {
         if !self.has_fields() {
             return Err(PyValueError::new_err(
                 "cannot get field information: type descriptor has no fields",
             ));
         }
-        let dict = unsafe { PyDict::from_borrowed_ptr(self.py(), (*self.as_dtype_ptr()).fields) };
+        let dict = unsafe {
+            Borrowed::from_ptr(self.py(), PyDataType_FIELDS(self.py(), self.as_dtype_ptr()))
+        };
+        let dict = unsafe { dict.downcast_unchecked::<PyDict>() };
         // NumPy guarantees that fields are tuples of proper size and type, so this should never panic.
         let tuple = dict
-            .get_item(name)
+            .get_item(name)?
             .ok_or_else(|| PyIndexError::new_err(name.to_owned()))?
-            .downcast::<PyTuple>()
+            .downcast_into::<PyTuple>()
             .unwrap();
         // Note that we cannot just extract the entire tuple since the third element can be a title.
-        let dtype = FromPyObject::extract(tuple.as_ref().get_item(0).unwrap()).unwrap();
-        let offset = FromPyObject::extract(tuple.as_ref().get_item(1).unwrap()).unwrap();
+        let dtype = tuple
+            .get_item(0)
+            .unwrap()
+            .downcast_into::<PyArrayDescr>()
+            .unwrap();
+        let offset = tuple.get_item(1).unwrap().extract().unwrap();
         Ok((dtype, offset))
     }
 }
 
+impl Sealed for Bound<'_, PyArrayDescr> {}
+
 /// Represents that a type can be an element of `PyArray`.
 ///
 /// Currently, only integer/float/complex/object types are supported. The [NumPy documentation][enumerated-types]
 /// list the other built-in types which we are not yet implemented.
 ///
+/// Note that NumPy's integer types like `numpy.int_` and `numpy.uint` are based on C's integer hierarchy
+/// which implies that their widths change depending on the platform's [data model][data-models].
+/// For example, `numpy.int_` matches C's `long` which is 32 bits wide on Windows (using the LLP64 data model)
+/// but 64 bits wide on Linux (using the LP64 data model).
+///
+/// In contrast, Rust's [`isize`] and [`usize`] types are defined to have the same width as a pointer
+/// and are therefore always 64 bits wide on 64-bit platforms. If you want to match NumPy's behaviour,
+/// consider using the [`c_long`][std::ffi::c_long] and [`c_ulong`][std::ffi::c_ulong] type aliases.
+///
 /// # Safety
 ///
 /// A type `T` that implements this trait should be safe when managed by a NumPy
@@ -386,7 +492,8 @@ impl PyArrayDescr {
 /// safely and efficiently using [`from_owned_object_array`][crate::PyArray::from_owned_object_array].
 ///
 /// [enumerated-types]: https://numpy.org/doc/stable/reference/c-api/dtype.html#enumerated-types
-pub unsafe trait Element: Clone + Send {
+/// [data-models]: https://en.wikipedia.org/wiki/64-bit_computing#64-bit_data_models
+pub unsafe trait Element: Sized + Send + Sync {
     /// Flag that indicates whether this type is trivially copyable.
     ///
     /// It should be set to true for all trivially copyable types (like scalar types
@@ -397,7 +504,41 @@ pub unsafe trait Element: Clone + Send {
     const IS_COPY: bool;
 
     /// Returns the associated type descriptor ("dtype") for the given element type.
-    fn get_dtype(py: Python) -> &PyArrayDescr;
+    fn get_dtype(py: Python<'_>) -> Bound<'_, PyArrayDescr>;
+
+    /// Deprecated name for [`Element::get_dtype`].
+    #[deprecated(since = "0.23.0", note = "renamed to `Element::get_dtype`")]
+    #[inline]
+    fn get_dtype_bound(py: Python<'_>) -> Bound<'_, PyArrayDescr> {
+        Self::get_dtype(py)
+    }
+
+    /// Create a clone of the value while the GIL is guaranteed to be held.
+    fn clone_ref(&self, py: Python<'_>) -> Self;
+
+    /// Create an owned copy of the slice while the GIL is guaranteed to be held.
+    ///
+    /// Some types may provide implementations of this method that are more efficient
+    /// than simply mapping the `py_clone` method to each element in the slice.
+    #[inline]
+    fn vec_from_slice(py: Python<'_>, slc: &[Self]) -> Vec<Self> {
+        slc.iter().map(|elem| elem.clone_ref(py)).collect()
+    }
+
+    /// Create an owned copy of the array while the GIL is guaranteed to be held.
+    ///
+    /// Some types may provide implementations of this method that are more efficient
+    /// than simply mapping the `py_clone` method to each element in the view.
+    #[inline]
+    fn array_from_view<D>(
+        py: Python<'_>,
+        view: ::ndarray::ArrayView<'_, Self, D>,
+    ) -> ::ndarray::Array<Self, D>
+    where
+        D: ::ndarray::Dimension,
+    {
+        view.map(|elem| elem.clone_ref(py))
+    }
 }
 
 fn npy_int_type_lookup<T, T0, T1, T2>(npy_types: [NPY_TYPES; 3]) -> NPY_TYPES {
@@ -414,7 +555,7 @@ fn npy_int_type_lookup<T, T0, T1, T2>(npy_types: [NPY_TYPES; 3]) -> NPY_TYPES {
 
 fn npy_int_type<T: Bounded + Zero + Sized + PartialEq>() -> NPY_TYPES {
     let is_unsigned = T::min_value() == T::zero();
-    let bit_width = size_of::<T>() << 3;
+    let bit_width = 8 * size_of::<T>();
 
     match (is_unsigned, bit_width) {
         (false, 8) => NPY_TYPES::NPY_BYTE,
@@ -445,14 +586,46 @@ fn npy_int_type<T: Bounded + Zero + Sized + PartialEq>() -> NPY_TYPES {
     }
 }
 
+// Invoke within the `Element` impl for a `Clone` type to provide an efficient
+// implementation of the cloning methods
+macro_rules! clone_methods_impl {
+    ($Self:ty) => {
+        #[inline]
+        fn clone_ref(&self, _py: ::pyo3::Python<'_>) -> $Self {
+            ::std::clone::Clone::clone(self)
+        }
+
+        #[inline]
+        fn vec_from_slice(_py: ::pyo3::Python<'_>, slc: &[$Self]) -> Vec<$Self> {
+            ::std::borrow::ToOwned::to_owned(slc)
+        }
+
+        #[inline]
+        fn array_from_view<D>(
+            _py: ::pyo3::Python<'_>,
+            view: ::ndarray::ArrayView<'_, $Self, D>,
+        ) -> ::ndarray::Array<$Self, D>
+        where
+            D: ::ndarray::Dimension,
+        {
+            ::ndarray::ArrayView::to_owned(&view)
+        }
+    };
+}
+pub(crate) use clone_methods_impl;
+use pyo3::BoundObject;
+
 macro_rules! impl_element_scalar {
     (@impl: $ty:ty, $npy_type:expr $(,#[$meta:meta])*) => {
         $(#[$meta])*
         unsafe impl Element for $ty {
             const IS_COPY: bool = true;
-            fn get_dtype(py: Python) -> &PyArrayDescr {
+
+            fn get_dtype(py: Python<'_>) -> Bound<'_, PyArrayDescr> {
                 PyArrayDescr::from_npy_type(py, $npy_type)
             }
+
+            clone_methods_impl!($ty);
         }
     };
     ($ty:ty => $npy_type:ident $(,#[$meta:meta])*) => {
@@ -471,6 +644,27 @@ impl_element_scalar!(u8, u16, u32, u64);
 impl_element_scalar!(f32 => NPY_FLOAT);
 impl_element_scalar!(f64 => NPY_DOUBLE);
 
+#[cfg(feature = "half")]
+impl_element_scalar!(f16 => NPY_HALF);
+
+#[cfg(feature = "half")]
+unsafe impl Element for bf16 {
+    const IS_COPY: bool = true;
+
+    fn get_dtype(py: Python<'_>) -> Bound<'_, PyArrayDescr> {
+        static DTYPE: GILOnceCell<Py<PyArrayDescr>> = GILOnceCell::new();
+
+        DTYPE
+            .get_or_init(py, || {
+                PyArrayDescr::new(py, "bfloat16").expect("A package which provides a `bfloat16` data type for NumPy is required to use the `half::bf16` element type.").unbind()
+            })
+            .clone_ref(py)
+            .into_bound(py)
+    }
+
+    clone_methods_impl!(Self);
+}
+
 impl_element_scalar!(Complex32 => NPY_CFLOAT,
     #[doc = "Complex type with `f32` components which maps to `numpy.csingle` (`numpy.complex64`)."]);
 impl_element_scalar!(Complex64 => NPY_CDOUBLE,
@@ -482,43 +676,53 @@ impl_element_scalar!(usize, isize);
 unsafe impl Element for PyObject {
     const IS_COPY: bool = false;
 
-    fn get_dtype(py: Python) -> &PyArrayDescr {
+    fn get_dtype(py: Python<'_>) -> Bound<'_, PyArrayDescr> {
         PyArrayDescr::object(py)
     }
+
+    #[inline]
+    fn clone_ref(&self, py: Python<'_>) -> Self {
+        Py::clone_ref(self, py)
+    }
 }
 
 #[cfg(test)]
 mod tests {
     use super::*;
 
-    use pyo3::py_run;
+    use pyo3::types::PyString;
+    use pyo3::{py_run, types::PyTypeMethods};
 
-    use crate::npyffi::NPY_NEEDS_PYAPI;
+    use crate::npyffi::{is_numpy_2, NPY_NEEDS_PYAPI};
 
     #[test]
     fn test_dtype_new() {
         Python::with_gil(|py| {
             assert!(PyArrayDescr::new(py, "float64")
                 .unwrap()
-                .is(dtype::<f64>(py)));
+                .is(&dtype::<f64>(py)));
 
             let dt = PyArrayDescr::new(py, [("a", "O"), ("b", "?")].as_ref()).unwrap();
-            assert_eq!(dt.names(), Some(vec!["a", "b"]));
+            assert_eq!(dt.names(), Some(vec!["a".to_owned(), "b".to_owned()]));
             assert!(dt.has_object());
-            assert!(dt.get_field("a").unwrap().0.is(dtype::<PyObject>(py)));
-            assert!(dt.get_field("b").unwrap().0.is(dtype::<bool>(py)));
+            assert!(dt.get_field("a").unwrap().0.is(&dtype::<PyObject>(py)));
+            assert!(dt.get_field("b").unwrap().0.is(&dtype::<bool>(py)));
 
-            assert!(PyArrayDescr::new(py, &123_usize).is_err());
+            assert!(PyArrayDescr::new(py, 123_usize).is_err());
         });
     }
 
     #[test]
     fn test_dtype_names() {
-        fn type_name<T: Element>(py: Python) -> &str {
-            dtype::<T>(py).typeobj().name().unwrap()
+        fn type_name<T: Element>(py: Python<'_>) -> Bound<'_, PyString> {
+            dtype::<T>(py).typeobj().qualname().unwrap()
         }
         Python::with_gil(|py| {
-            assert_eq!(type_name::<bool>(py), "bool_");
+            if is_numpy_2(py) {
+                assert_eq!(type_name::<bool>(py), "bool");
+            } else {
+                assert_eq!(type_name::<bool>(py), "bool_");
+            }
 
             assert_eq!(type_name::<i8>(py), "int8");
             assert_eq!(type_name::<i16>(py), "int16");
@@ -555,7 +759,7 @@ mod tests {
 
             assert_eq!(dt.num(), NPY_TYPES::NPY_DOUBLE as c_int);
             assert_eq!(dt.flags(), 0);
-            assert_eq!(dt.typeobj().name().unwrap(), "float64");
+            assert_eq!(dt.typeobj().qualname().unwrap(), "float64");
             assert_eq!(dt.char(), b'd');
             assert_eq!(dt.kind(), b'f');
             assert_eq!(dt.byteorder(), b'=');
@@ -563,13 +767,13 @@ mod tests {
             assert_eq!(dt.itemsize(), 8);
             assert_eq!(dt.alignment(), 8);
             assert!(!dt.has_object());
-            assert_eq!(dt.names(), None);
+            assert!(dt.names().is_none());
             assert!(!dt.has_fields());
             assert!(!dt.is_aligned_struct());
             assert!(!dt.has_subarray());
-            assert!(dt.base().is_equiv_to(dt));
+            assert!(dt.base().is_equiv_to(&dt));
             assert_eq!(dt.ndim(), 0);
-            assert_eq!(dt.shape(), vec![]);
+            assert_eq!(dt.shape(), Vec::<usize>::new());
         });
     }
 
@@ -585,12 +789,13 @@ mod tests {
             let dt = locals
                 .get_item("dtype")
                 .unwrap()
-                .downcast::<PyArrayDescr>()
+                .unwrap()
+                .downcast_into::<PyArrayDescr>()
                 .unwrap();
 
             assert_eq!(dt.num(), NPY_TYPES::NPY_VOID as c_int);
             assert_eq!(dt.flags(), 0);
-            assert_eq!(dt.typeobj().name().unwrap(), "void");
+            assert_eq!(dt.typeobj().qualname().unwrap(), "void");
             assert_eq!(dt.char(), b'V');
             assert_eq!(dt.kind(), b'V');
             assert_eq!(dt.byteorder(), b'|');
@@ -598,13 +803,13 @@ mod tests {
             assert_eq!(dt.itemsize(), 48);
             assert_eq!(dt.alignment(), 8);
             assert!(!dt.has_object());
-            assert_eq!(dt.names(), None);
+            assert!(dt.names().is_none());
             assert!(!dt.has_fields());
             assert!(!dt.is_aligned_struct());
             assert!(dt.has_subarray());
             assert_eq!(dt.ndim(), 2);
             assert_eq!(dt.shape(), vec![2, 3]);
-            assert!(dt.base().is_equiv_to(dtype::<f64>(py)));
+            assert!(dt.base().is_equiv_to(&dtype::<f64>(py)));
         });
     }
 
@@ -620,14 +825,15 @@ mod tests {
             let dt = locals
                 .get_item("dtype")
                 .unwrap()
-                .downcast::<PyArrayDescr>()
+                .unwrap()
+                .downcast_into::<PyArrayDescr>()
                 .unwrap();
 
             assert_eq!(dt.num(), NPY_TYPES::NPY_VOID as c_int);
             assert_ne!(dt.flags() & NPY_ITEM_HASOBJECT, 0);
             assert_ne!(dt.flags() & NPY_NEEDS_PYAPI, 0);
             assert_ne!(dt.flags() & NPY_ALIGNED_STRUCT, 0);
-            assert_eq!(dt.typeobj().name().unwrap(), "void");
+            assert_eq!(dt.typeobj().qualname().unwrap(), "void");
             assert_eq!(dt.char(), b'V');
             assert_eq!(dt.kind(), b'V');
             assert_eq!(dt.byteorder(), b'|');
@@ -635,21 +841,24 @@ mod tests {
             assert_eq!(dt.itemsize(), 24);
             assert_eq!(dt.alignment(), 8);
             assert!(dt.has_object());
-            assert_eq!(dt.names(), Some(vec!["x", "y", "z"]));
+            assert_eq!(
+                dt.names(),
+                Some(vec!["x".to_owned(), "y".to_owned(), "z".to_owned()])
+            );
             assert!(dt.has_fields());
             assert!(dt.is_aligned_struct());
             assert!(!dt.has_subarray());
             assert_eq!(dt.ndim(), 0);
-            assert_eq!(dt.shape(), vec![]);
-            assert!(dt.base().is_equiv_to(dt));
+            assert_eq!(dt.shape(), Vec::<usize>::new());
+            assert!(dt.base().is_equiv_to(&dt));
             let x = dt.get_field("x").unwrap();
-            assert!(x.0.is_equiv_to(dtype::<u8>(py)));
+            assert!(x.0.is_equiv_to(&dtype::<u8>(py)));
             assert_eq!(x.1, 0);
             let y = dt.get_field("y").unwrap();
-            assert!(y.0.is_equiv_to(dtype::<f64>(py)));
+            assert!(y.0.is_equiv_to(&dtype::<f64>(py)));
             assert_eq!(y.1, 8);
             let z = dt.get_field("z").unwrap();
-            assert!(z.0.is_equiv_to(dtype::<PyObject>(py)));
+            assert!(z.0.is_equiv_to(&dtype::<PyObject>(py)));
             assert_eq!(z.1, 16);
         });
     }
